--- Repository interface module.
-- Holds information of game elements like tracks, streets, bridges, tunnels, crossings
-- @author Oskar Eisemuth
-- @copyright 2017-2019
-- @module commonapi.repos
local repos = {}

local _log = commonapi.utils.logPrep("commonapi2.repos")
local debugrepo = true
local repoNames = { 
	"street", 
	"track", 
	"bridge", 
	"tunnel", 
	"railroadCrossing"
}

local repoFolders = { 
	"street", 
	"track", 
	"bridge", 
	"tunnel", 
	"railroad_crossing"
}

local repoFolderMap = { 
	street = "street", 
	track = "track", 
	bridge = "bridge", 
	tunnel = "tunnel", 
	railroad_crossing = "railroadCrossing"
}

local streetenv = "default"
local streetfilter = {}

streetfilter.default = function(fileName)
	local subdir = string.match(fileName, "([^%/]+)/.+[.]lua$")
	if (subdir == nil) then
		return false
	end
	if (subdir == "usa") then
		return true
	end
	if (subdir == "airport") then
		return true
	end
	return false
end

streetfilter.usa = function(fileName)
	local subdir = string.match(fileName, "([^%/]+)/.+[.]lua$")
--[[
	if (subdir == nil) then
		if fileName == "country_new_small.lua" then return true end
		if fileName == "country_new_medium.lua" then return true end
		if fileName == "country_new_medium_upgrade.lua" then return true end
		if fileName == "country_new_large.lua" then return true end
		if fileName == "country_new_large_upgrade.lua" then return true end
		if fileName == "country_new_x_large.lua" then return true end
		if fileName == "country_new_x_large_upgrade.lua" then return true end
		if fileName == "new_small.lua" then return true end
		return false
	end
	if (subdir == "usa") then
		return true
	end
--]]
	if (subdir == "airport") then
		return true
	end
	return false
end


local function prepStorage(filelist)
	local utils = commonapi.utils
	local modiferRegex = ".*/res/config/([^%/]+)/(.+[.]lua)$"
	local modidRegex = "([^%/]+)/([^%/]+)/res/config/[^%/]+/.+[.]lua$"
	
	local pathRegex = "^(.+)/res/config/[^%/]+/.+[.]lua$"


	utils._detectLanguage()
	
	local streetfilterfnc = streetfilter['default']
	if (streetfilter[streetenv] ~= nil) then
		streetfilterfnc = streetfilter[streetenv]
	end
	
	_log("Preloading "..#filelist.." Files")
	for d = 1, #filelist do
		local filename = filelist[d]
		local repodir, basename = string.match(filename, modiferRegex)
		
		if (repodir == "street") then
			if (streetfilterfnc(basename)) then
				if (debugrepo) then
					_log("Filtered "..basename)
				end
				basename = nil
			end
		end
		
		if (repodir ~= nil and basename ~= nil) then
			local repoName = repoFolderMap[repodir]
			if (repoName ~= nil) then
				local translatemodid = "_unknown_modid_"
				local translateString = utils.translateString

				local folder, modidr = string.match(filename, modidRegex)
				if (folder == "mods") then
					translatemodid = modidr
				elseif (folder == "1066780") then
					translatemodid = "*"..modidr.."_1"
				elseif (folder == "staging_area") then
					translatemodid = "?"..modidr
				end
				local prefolder = string.match(filename, pathRegex)



				local env = {
					data = function()
						return {}
					end,
					["_"] = function(s)
						return translateString(translatemodid, s);
					end
				}

				local oldpackagepath = package.path
				package.path = prefolder.."/res/scripts/?.lua;" ..oldpackagepath

				local data = utils.getFileData(filename, env)
				if (data ~= nil) then
					if (debugrepo) then
						_log("Preloading "..repoName.." "..basename)
					end

					data = applyFileFilters(repoName, filename, data)
					if (data ~= nil) then 
						repos[repoName]._modifier(filename, data)
					else
						if (debugrepo) then
							_log("Filtered "..repoName.." "..basename)
						end
					end
				end
				package.path = oldpackagepath
			end
		end
	end
	_log("Preloading done")
end

local function findModContent()
	local mods = commonapi.mods.getLoaded()
	
	local dirs = {}
	
	for i = 1, #repoNames do
		local line = "res/config/"..repoFolders[i].."/"
		dirs[#dirs+1] = line
	end
		
	for m, entry in pairs( mods ) do
		for r = 1, #repoNames do
			local line = entry["path"].."res/config/"..repoFolders[r].."/"
			dirs[#dirs+1] = line
		end
	end
	local filelist = commonapi.utils.scanDirs(dirs)
	
	prepStorage(filelist)
end


--- Repository Object.
-- Object accessed by repository interface 
-- @type Repository

--- Repository object factory.
-- @local 
-- @treturn repoobj
local function createRepository(typ)

	local repoobj = {}
	local repotyp = typ
	
	local storage = {}
	local idmap = {}

	local shared_key = "repository_"..typ
	local shared_serial = nil
	local _shareddataptr

	if (commonapi._native ~= nil and commonapi._native.shareddata ~= nil) then
		_shareddataptr = commonapi._native.shareddata
	end

	local modiferRegex = "res/config/([^%/]+)/(.+[.]lua)$"
	
	local modidRegex = "([^%/]+)/([^%/]+)/res/config/[^%/]+/.+[.]lua$"

	-- Private API --

	repoobj._getSyncStorage = function()
		if (_shareddataptr == nil) then
			return
		end
		local newserial =_shareddataptr.getSerial(shared_key)
		if (newserial ~= nil) then
			if (newserial ~= shared_serial) then
				_log("Syncing shareddata ("..shared_key..") new: ".. (newserial or 'nil ') .. " old: " .. (shared_serial or 'nil ') .. " ".. commonapi._native.getLuaStatePtr() )
				local newdata = _shareddataptr.get(shared_key)
				storage = newdata['storage'] or {}
				idmap = newdata['idmap'] or {}
			end
		end
	end

	repoobj._doSyncStorage = function()
		if (_shareddataptr == nil) then
			return
		end
		local newdata =  {
			["idmap"] = idmap,
			["storage"] = storage
		}
		shared_serial = _shareddataptr.set(shared_key, newdata)
	end

	repoobj._modifier = function(fileName, data)
		repoobj._getSyncStorage()

		local repodir, basename = string.match(fileName, modiferRegex)
		
		if (basename == nil) then
			return data
		end
		
		if (type(data) ~= "table") then
			storage[basename] = nil
			return data
		end

		local entry = storage[basename]
		if (type(entry) ~= "table") then
			local id = #idmap+1
			
			entry = {
				filename = basename,
				id = id,
				data = data,
			}
			idmap[id] = basename
			storage[basename] = entry
			
			-- Fix high_speed.lua/standard.lua
			if (id == 2 and basename == "standard.lua") then
				if (idmap[1] == "high_speed.lua") then
					idmap[1] = "standard.lua"
					idmap[2] = "high_speed.lua"
				end
			end
			local folder, modid = string.match(fileName, modidRegex)
			if (folder == "mods" or folder == "1066780") then
				entry.modid = modid
			end
		else 
			-- the modifer gets called by prepStorage, so data table need to be fixed to later point to the game data table
			entry.data = data
		end
		
		repoobj._doSyncStorage()
		return data
	end		
		
	repoobj._addModifer = function()
		local modiferName = "load"..repotyp:sub(1,1):upper()..repotyp:sub(2)
		addModifier(modiferName, repoobj._modifier)
	end
	
	-- Public API --
	
	--- Get count of Entries.
	repoobj.getCount = function()
		repoobj._getSyncStorage()
		return #idmap
	end
	
	--- Return entries.
	-- @treturn {repoentry,...}
	repoobj.getEntries = function()
		repoobj._getSyncStorage()
		local entries = {}
		local entry
		
		for i, basename in ipairs(idmap) do
			entry = storage[basename]
			if (entry ~= nil) then
				entries[#entries+1] = table.copy(entry)
			end
		end
		return entries
	end
	
	--- Return filteded entries.
	-- @tparam {function,...} filters A list of filters to apply
	-- @treturn {repoentry,...}
	repoobj.getEntriesFiltered = function(filters)
		repoobj._getSyncStorage()
		local entries = {}
		local entry, entryok
		
		for i, basename in ipairs(idmap) do
			entry = storage[basename]
			if (entry ~= nil) then
				entryok = true
				for n, fn in ipairs(filters) do
					if not fn(entry.filename, entry.data) then
						entryok = false
						break
					end
				end
				if (entryok) then
					entries[#entries+1] = table.copy(entry)
				end
			end
		end
		return entries
	end
	
	--- Return entry by id.
	-- @tparam integer id Id of entry
	-- @treturn repoentry Entry
	repoobj.getById = function(id)
		repoobj._getSyncStorage()
		local basename = idmap[id]
		if (basename ~= nil) then 
			local entry = storage[basename]
			if (type(entry) == "table") then
				assert(entry["id"] == id, "repoobj("..repotyp..")getById mapping mismatch for "..basename)
				return table.copy(entry)
			end
		end
		return nil
	end
	
	--- Return entry by name.
	-- @tparam string name Name of entry
	-- @treturn repoentry Entry
	repoobj.getByName = function(name)
		repoobj._getSyncStorage()
		local entry = storage[name]
		if (type(entry) == "table") then
			return table.copy(entry)
		end
		return nil
	end
	
	----
	return repoobj
end


---- Repository Interface.
-- Public interface to get specific repositories.
-- @section repoaccess

----
--- modifer support glue (internal)
-- @local 
repos._addModifer = function()
	commonapi.utils._detectLanguage()
	
	for i = 1, #repoNames do
		local name = repoNames[i]
		repos[name]._addModifer()
	end
end

repos._findModContentGlue = findModContent
repos._prepSettings = function(settings)
	streetenv = settings.streets

	local ok, laneutil = pcall(require, "laneutil")
	if (ok) then
		local makeLanesScale = commonapi.utils._getFunctionLocal(laneutil.makeLanes, "scale")
		if (type(makeLanesScale) == "function") then
			_log("Patching laneutil.makeLanes")
			local laneutilfix = require("commonapi2.fix.laneutilfix")
			laneutil.makeLanes = laneutilfix.makeLanes
		end
	end
end


-- Does not work with ldoc
--[[
local function initRepositories()
	for i = 1, #repoNames do
		local name = repoNames[i]
		repos[name] = createRepository(name)
	end
end
initRepositories()
--]]

--- Track repository
repos.track = createRepository("track")

--- Street repository
repos.street = createRepository("street")

--- Bridge repository
repos.bridge = createRepository("bridge")

--- Tunnel repository
repos.tunnel = createRepository("tunnel")

--- Railroad crossing repository
-- @see Repository
repos.railroadCrossing = createRepository("railroadCrossing")


----
--- Repos Types
-- @section reposstypes

---- A repo entry.
-- Item returned by repository functions
-- @table repoentry
---@field filename filename (without path)
-- @field id internal id
-- @field data Data or nil

return repos
