local ModSettingsWindow = {}

local _log = commonapi.utils.logPrep("commonapi2.ui.ModSettingsWindow")
local _utils = commonapi.utils
local _ = _utils._t
local ui = commonapi.ui;

ModSettingsWindow = ui.Window:new()

ModSettingsWindow.id =  "ModSettingsWindow"
ModSettingsWindow.uilist = {}

ModSettingsWindow.height = 500
ModSettingsWindow.width = 500

function ModSettingsWindow.create(modid)
	local modentry = commonapi.mods.getById(modid)
    return ModSettingsWindow.createByModEntry(modentry)
end

function ModSettingsWindow.createByModEntry(modentry)
	local o
	local modid = modentry.id
	o = ModSettingsWindow:new()
	o.id = "ModSettingsWindow_"..modid
	o.caption = modentry.name.."  ("..modid..")"
	o.modid = modid

	commonapi.settings._clearCache(modid)

	local settingsdesc = commonapi.settings._getDescriptionFromEntry(modentry)
	local settingsorder = {}

	if (settingsdesc ~= nil) then
		for key, desc in pairs(settingsdesc) do
			local order = desc.order or 0
			settingsorder[#settingsorder+1] = {key, order}
		end

		table.sort(settingsorder,
			function(a, b)
				return a[2] < b[2]
			end
		)
	end

	o.data = {
		entry = modentry,
		settingsdesc = settingsdesc,
		settingsorder = settingsorder,
		oldsettings = commonapi.settings._getFromEntry(modentry)
	}

	o:createUIList()
	return o
end


function ModSettingsWindow:createUIList()
	local uilist = {}

	local settings =  self.data.oldsettings
	local settingsdesc = self.data.settingsdesc
	local settingsorder = self.data.settingsorder
	if (settingsdesc == nil) then
		self.uilist ={
			{
				typ = "Label",
				caption = "No Settings found"
			}
		}
		return
	end
	
	uilist[#uilist+1] = {
		typ = "_ScrollRegionStart",
		id = "ScrollRegion",
		border = false,
		size = {
			0, -30
		}
	}
	--for key, desc in pairs(settingsdesc) do
	for _, orderdata in ipairs(settingsorder) do
		local key = orderdata[1]
		local desc = settingsdesc[key]

		uilist[#uilist+1] = {
			typ = "Label",
			caption = desc.name
		}

		if (desc.description ~= nil) then
			uilist[#uilist+1] = {
				typ = "Label",
				caption = desc.description
			}
		end
		if (desc.type == "table") then
			for tkey, tvalue in ipairs(desc.values) do
				local tvalueid = key..'_'..tvalue.value
				uilist[#uilist+1] = {
					typ = "Checkbox",
					id = tvalueid,
					caption = ''..(tvalue.text or tvalue.value)
				}
			end
			if (type(settings[key]) == 'table') then
				for tkey, pvalue in ipairs(settings[key]) do
					self.params[key..'_'..pvalue] = true
				end
			end
		else
			if (desc.type == "number") then
				uilist[#uilist+1] = {
					typ = "FloatValue",
					id = key,
				}
				if (desc.subtype == "integer") then
					uilist[#uilist].typ = "IntValue"
				end
			elseif (desc.type == "string") then
				uilist[#uilist+1] = {
					typ = "TextInput",
					id = key,
				}
			elseif (desc.type == "boolean") then
				uilist[#uilist+1] = {
					typ = "Checkbox",
					id = key,
				}
			end
			self.params[key] = settings[key] or nil
		end
		
		uilist[#uilist+1] = {
			typ = "Separator"
		}
	end

	uilist[#uilist+1] = {
		typ = "_ScrollRegionEnd",
		id = "ScrollRegionEnd"
	}

	uilist[#uilist+1] = {
		typ = "Separator"
	}
	uilist[#uilist+1] = {
		typ = "Button",
		id = "SaveSettings",
		caption = _("Save"),
		onclick = function(item, win)
			win:saveSetting()
		end
	}

	self.uilist = uilist
end

local function serializeSettingsData(o, prefix)
	local typ = type(o)
	if typ == "nil" then
		return "nil"
	elseif typ == "boolean" then
		return tostring(o)
	elseif typ == "number" then
		return tostring(o)
	elseif typ == "string" then
		return string.format("%q", o)
	elseif typ == "table" then
		local newprefix = prefix.."\t"
		local seenindex = {}
		local ret = {}

		ret[#ret+1] = "{\n"
		ret[#ret+1] = newprefix
		for k, v in ipairs(o) do
			seenindex[k] = true
			ret[#ret+1] = serializeSettingsData(v, newprefix)
			ret[#ret+1] = ", "
		end

		if (#seenindex > 0) then
			ret[#ret+1] = "\n"
			ret[#ret+1] = newprefix
		end

		for k, v in pairs(o) do
			if (seenindex[k] == nil) then
				local ktype = type(k)
				if (ktype == "string" and string.find(k, "^[_%a][_%w]*$")) then
					ret[#ret+1] = k
					ret[#ret+1] = " = "
				else
					ret[#ret+1] = "["
					serializeSettingsData(k, newprefix)
					ret[#ret+1] = "] = "
				end
				ret[#ret+1] = serializeSettingsData(v, newprefix)
				ret[#ret+1] = ",\n"
				ret[#ret+1] = newprefix
			end
		end
		ret[#ret+1] = "\n"
		ret[#ret+1] = prefix
		ret[#ret+1] = "}"
		return table.concat(ret)
	else
		error("serializeSettingsData can't serialize " .. typ)
	end
end

ModSettingsWindow.serializeSettingsData = serializeSettingsData

function ModSettingsWindow:saveSetting()
	local settingsdesc = self.data.settingsdesc
	local newsettings = {}

	for key, desc in pairs(settingsdesc) do
		local value = self.params[key]
		if (desc.type == "number") then
			newsettings[key] = self.params[key]
		elseif (desc.type == "string") then
			newsettings[key] = ''.. (self.params[key] or '')
		elseif (desc.type == "boolean") then
			newsettings[key] = (self.params[key] or false)
		elseif (desc.type == "table") then
			local newtable = {}
			for tkey, tvalue in ipairs(desc.values) do
				local tvalueid = key..'_'..tvalue.value

				if (self.params[tvalueid] == true) then
					newtable[#newtable+1] = tvalue.value
				end
			end
			newsettings[key] = newtable
		end
	end

	local entry = self.data.entry

	if (entry ~= nil and entry.path ~= nil and entry.data ~= nil) then
		local filename = entry.path .. "settings.lua"
		local serData = "return "..serializeSettingsData(newsettings, "").."\n"
		local fh = io.open(filename, "w")
		if (fh) then
			fh:write(serData)
			fh:close()
		end
		commonapi.settings._clearCache(self.modid)
	end

	self:close()
end


return ModSettingsWindow